@@ -25,6 +25,8 @@
 #include "debug.h"
 #include <QMutexLocker>
 #include <QTimer>
+#include <QDBusServiceWatcher>
+#include <QDBusConnection>
 
 #include "card.h"
 #include "client.h"
@@ -33,6 +35,7 @@
 #include "source.h"
 #include "sourceoutput.h"
 #include "streamrestore.h"
+#include "module.h"
 
 namespace QPulseAudio
 {
@@ -130,6 +133,15 @@ static void card_cb(pa_context *context, const pa_card_info *info, int eol, void
     ((Context *)data)->cardCallback(info);
 }
 
+static void module_info_list_cb(pa_context *context, const pa_module_info *info, int eol, void *data)
+{
+    if (!isGoodState(eol))
+        return;
+    Q_ASSERT(context);
+    Q_ASSERT(data);
+    ((Context *)data)->moduleCallback(info);
+}
+
 static void server_cb(pa_context *context, const pa_server_info *info, void *data)
 {
     Q_ASSERT(context);
@@ -168,6 +180,46 @@ static void ext_stream_restore_subscribe_cb(pa_context *context, void *data)
     }
 }
 
+static void ext_stream_restore_change_sink_cb(pa_context *context, const pa_ext_stream_restore_info *info, int eol, void *data)
+{
+    if (!isGoodState(eol)) {
+        return;
+    }
+    Q_ASSERT(context);
+    Q_ASSERT(data);
+    if (qstrncmp(info->name, "sink-input-by", 13) == 0) {
+        Context *context = static_cast<Context *>(data);
+        const QByteArray deviceData = context->newDefaultSink().toUtf8();
+        pa_ext_stream_restore_info newinfo;
+        newinfo.name = info->name;
+        newinfo.channel_map = info->channel_map;
+        newinfo.volume = info->volume;
+        newinfo.mute = info->mute;
+        newinfo.device = deviceData.constData();
+        context->streamRestoreWrite(&newinfo);
+    }
+}
+
+static void ext_stream_restore_change_source_cb(pa_context *context, const pa_ext_stream_restore_info *info, int eol, void *data)
+{
+    if (!isGoodState(eol)) {
+        return;
+    }
+    Q_ASSERT(context);
+    Q_ASSERT(data);
+    if (qstrncmp(info->name, "source-output-by", 16) == 0) {
+        Context *context = static_cast<Context *>(data);
+        const QByteArray deviceData = context->newDefaultSource().toUtf8();
+        pa_ext_stream_restore_info newinfo;
+        newinfo.name = info->name;
+        newinfo.channel_map = info->channel_map;
+        newinfo.volume = info->volume;
+        newinfo.mute = info->mute;
+        newinfo.device = deviceData.constData();
+        context->streamRestoreWrite(&newinfo);
+    }
+}
+
 // --------------------------
 
 Context::Context(QObject *parent)
@@ -177,6 +229,16 @@ Context::Context(QObject *parent)
     , m_mainloop(nullptr)
     , m_references(0)
 {
+    QDBusServiceWatcher *watcher = new QDBusServiceWatcher(QStringLiteral("org.pulseaudio.Server"),
+                                                           QDBusConnection::sessionBus(),
+                                                           QDBusServiceWatcher::WatchForRegistration,
+                                                           this);
+    connect(watcher, &QDBusServiceWatcher::serviceRegistered, this, [this]() {
+        if (!m_context) {
+            connectToDaemon();
+        }
+    });
+
     connectToDaemon();
 }
 
@@ -287,6 +349,18 @@ void Context::subscribeCallback(pa_context *context, pa_subscription_event_type_
         }
         break;
 
+    case PA_SUBSCRIPTION_EVENT_MODULE:
+        if ((type & PA_SUBSCRIPTION_EVENT_TYPE_MASK) == PA_SUBSCRIPTION_EVENT_REMOVE) {
+            m_modules.removeEntry(index);
+        } else {
+            if (!PAOperation(pa_context_get_module_info_list(context, module_info_list_cb, this))) {
+                qCWarning(PLASMAPA) << "pa_context_get_module_info_list() failed";
+                return;
+            }
+        }
+        break;
+
+
     case PA_SUBSCRIPTION_EVENT_SERVER:
         if (!PAOperation(pa_context_get_server_info(context, server_cb, this))) {
             qCWarning(PLASMAPA) << "pa_context_get_server_info() failed";
@@ -315,6 +389,7 @@ void Context::contextStateCallback(pa_context *c)
                                             PA_SUBSCRIPTION_MASK_SINK_INPUT|
                                             PA_SUBSCRIPTION_MASK_SOURCE_OUTPUT|
                                             PA_SUBSCRIPTION_MASK_CARD|
+                                            PA_SUBSCRIPTION_MASK_MODULE|
                                             PA_SUBSCRIPTION_MASK_SERVER), nullptr, nullptr))) {
                 qCWarning(PLASMAPA) << "pa_context_subscribe() failed";
                 return;
@@ -351,6 +426,11 @@ void Context::contextStateCallback(pa_context *c)
             return;
         }
 
+        if (!PAOperation(pa_context_get_module_info_list(c, module_info_list_cb, this))) {
+            qCWarning(PLASMAPA) << "pa_context_get_module_info_list() failed";
+            return;
+        }
+
         if (!PAOperation(pa_context_get_server_info(c, server_cb, this))) {
             qCWarning(PLASMAPA) << "pa_context_get_server_info() failed";
             return;
@@ -404,6 +484,11 @@ void Context::cardCallback(const pa_card_info *info)
     m_cards.updateEntry(info, this);
 }
 
+void Context::moduleCallback(const pa_module_info *info)
+{
+    m_modules.updateEntry(info, this);
+}
+
 void Context::streamRestoreCallback(const pa_ext_stream_restore_info *info)
 {
     if (qstrcmp(info->name, "sink-input-by-media-role:event") != 0) {
@@ -456,6 +541,14 @@ void Context::setDefaultSink(const QString &name)
                                                  nullptr))) {
         qCWarning(PLASMAPA) << "pa_context_set_default_sink failed";
     }
+
+    // Change device for all entries in stream-restore database
+    m_newDefaultSink = name;
+    if (!PAOperation(pa_ext_stream_restore_read(m_context,
+                                                ext_stream_restore_change_sink_cb,
+                                                this))) {
+        qCWarning(PLASMAPA) << "pa_ext_stream_restore_read failed";
+    }
 }
 
 void Context::setDefaultSource(const QString &name)
@@ -470,6 +563,14 @@ void Context::setDefaultSource(const QString &name)
                                                  nullptr))) {
         qCWarning(PLASMAPA) << "pa_context_set_default_source failed";
     }
+
+    // Change device for all entries in stream-restore database
+    m_newDefaultSource = name;
+    if (!PAOperation(pa_ext_stream_restore_read(m_context,
+                                                ext_stream_restore_change_source_cb,
+                                                this))) {
+        qCWarning(PLASMAPA) << "pa_ext_stream_restore_read failed";
+    }
 }
 
 void Context::streamRestoreWrite(const pa_ext_stream_restore_info *info)
@@ -526,6 +627,9 @@ void Context::reset()
     m_sources.reset();
     m_sourceOutputs.reset();
     m_clients.reset();
+    m_cards.reset();
+    m_modules.reset();
+    m_streamRestores.reset();
     m_server->reset();
 }
 
